/**
 * \file CoachKernel.cpp
 *
 * Kernel implementation of the CoachKernel class
 */

#include "Coach/CoachKernel.h"
#include "Common/Tasks.h"
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <limits.h>
#include <time.h>

#define MAX_PLAYER_AGE 15

///////////////////////////////////////////////////////////////////////////////
// CoachKernel interface

Interface* CoachKernel::GetInterface(int task) const
{
	if (task == MEMORY_TASK) return (CoachMemoryInterface*)this;
	else if (task == SYNCHRONIZATION_TASK) return (CoachSynchronizationInterface*)this;
	else return NULL;
}

///////////////////////////////////////////////////////////////////////////////
// CoachKernel construction

CoachKernel::CoachKernel(
	const char* name /* = "CoachKernel"*/,
	const char* desc /* = "CoachKernel module"*/)
	: Module(name, desc, "v0.0.001")
{
	// initialisation of simulation parameters
	serverParams = NULL;
	playerParams = NULL;
	playerTypes = NULL;

	playMode = PM_BeforeKickOff;
	timeInPlayOn = 0;
	continousTimeInPlayOn = 0;
	canSendCLmessage = 0;
	sayCoachCnt = 0;
	scoreOur = 0;
	scoreTheir = 0;
	timeOfLastVisualInfo = 0;
	memset(&lastBallInfo, 0, sizeof(lastBallInfo));
	lastNearestPlayer = 0;
	memset(&lastKick, 0, sizeof(lastKick));
	FIFObottom = FIFOtop = 0;
	advice = 0;
	nHeteroCounter=0;
	formationLastChangeTime = 0;
	formationTheirLastScore = 0;

	// all the players are of default type
	memset(&playerTypesByNumber, 0, sizeof(playerTypesByNumber));

	// no statistics by now
	memset(&statsBall, 0, sizeof(statsBall));
	memset(statsOurPlayer, 0, 12 * sizeof(statsOurPlayer[0]));
	memset(statsTheirPlayer, 0, 12 * sizeof(statsTheirPlayer[0]));

	// no spy information by now
	memset(spyTheirPlayer, 0, 12 * sizeof(PLAYER_SPY));
	memset(&spyBall, 0, sizeof(BALL_SPY));

	// no visual info dump by default
//	visualDumpFileName = NULL;
//	visualDumpFile = NULL;

  // logger
  log = new Log("coach-log");
	log->Init();
	log->Create(15);

	// default logo directory
	teamGraphicDir[0] = '\0';

	for(int i = 0; i < 32; i++) for(int j = 0; j < 8; j++) teamGraphicOk[i][j] = false;
	teamGraphicOkCount = 0;
	teamGraphicTries = 0;

	Main::GetInstance()->RegisterModule(this, MEMORY_TASK);
	Main::GetInstance()->RegisterModule(this, SYNCHRONIZATION_TASK);
////////////////////////////////////////////////////////////////
	for (i=0; i<12; i++)
		log->printf( "eZo08:%d speed(%f, %d), acceleration(%f, %d), rotation(%f, %d), speedXrotation(%f, %d)\n", \
				i, \
				spyTheirPlayer[i].speed.val, spyTheirPlayer[i].speed.time, \
				spyTheirPlayer[i].acceleration.val, spyTheirPlayer[i].acceleration.time,\
				spyTheirPlayer[i].rotation.val, spyTheirPlayer[i].rotation.time, \
				spyTheirPlayer[i].speedXRotation.val, spyTheirPlayer[i].speedXRotation.time);
////////////////////////////////////////////////////////////////

	statsPlayground = new PlaygroundStatistics(X_ZONES_NUMBER, Y_ZONES_NUMBER);

	currentFormation.InitializeFormation(INITIAL_FORMATION);
}

int CoachKernel::Init(int argc, char** argv, int)
{
	for(int i = 1; i < argc; i++)
	{
		if(strcmp(argv[i], "-h") == 0)
		{
			fprintf(stderr, "\n%s %s switches\n\n", (char*) GetName(), (char*) GetVer());
			fprintf(stderr, "\t-h: this help\n");
			fprintf(stderr, "\t-l <dir>: directory with team logo files\n\n");
			fprintf(stderr, "\t          <dir> HAS TO end with a slash/backslash, e.g.\n");
			fprintf(stderr, "\t          \"-l c:\\player\\logo\\\" or");
			fprintf(stderr, "\t          \"-l ~/player/logo\"");
			fprintf(stderr, "\t          \"--playground-stat-period Period to store the playground statistics to file (filname is 'PlaygroundStat_YYYYMMDDhhmm-SIMTIME.log')\"");

			return INIT_ERR;
		}
		else if(strcmp(argv[i], "-l") == 0)
		{
			if(argv[++i] != NULL)
				strncpy(teamGraphicDir, argv[i], 512);
		}
		else if (strcmp(argv[i], "--playground-stat-period") == 0)
		{
			if(argv[++i] != NULL)
				sscanf(argv[i], "%i", &statsPlaygroundStorePeriod);
		}
	}
/*
	/// opening of visual dump
	visualDumpFileName = "visualDump.bin";
	visualDumpFile = fopen( visualDumpFileName, "wb" );

	assert( visualDumpFile != NULL );
*/
	return INIT_OK;
}

///////////////////////////////////////////////////////////////////////////////
// CoachKernel operations
bool CoachKernel::IsTimeStoped() const
{
	switch(GetPlayMode())
	{
	// modes when time is stopped
	case PM_BeforeKickOff:
	case PM_AfterGoal_Our:
	case PM_AfterGoal_Their:
		return true;
	default:
		// simulation time is running
		return false;
	}
}

int CoachKernel::GetFormation(bool our)
{
	static float defs, mids, atts;
	static float min, lse; // least square error
	static int i, mini;

	atts = mids = defs = 0;
	if(our)
	{
		for(i = 1; i <= 11; i++)
		{
			atts += statsOurPlayer[i].cyclesAttacker;
			defs += statsOurPlayer[i].cyclesDefender;
			mids += statsOurPlayer[i].cyclesMidfielder;
		}
	}
	else
	{
		for(i = 1; i <= 11; i++)
		{
			atts += statsTheirPlayer[i].cyclesAttacker;
			defs += statsTheirPlayer[i].cyclesDefender;
			mids += statsTheirPlayer[i].cyclesMidfielder;
		}
	}
	defs = defs / timeInPlayOn;
	mids = mids / timeInPlayOn;
	atts = atts / timeInPlayOn;

	min = 1e10; // big number
	mini = 0;
	for(i = 0; i < NUM_FORMATIONS; i++)
	{
		lse = (formation[i].defs - defs) * (formation[i].defs - defs) +
			(formation[i].mids - mids) * (formation[i].mids - mids) +
			(formation[i].atts - atts) * (formation[i].atts - atts);
		if(lse < min)
		{
			min = lse;
			mini = i;
		}
	}

	return mini;
}

int CoachKernel::CanSendMessage()
{
	if(canSendCLmessage == 0) return 1;
	else // we can use freeform message (???)
	{
		static int ffwp = serverParams->freeform_wait_period;
		// number 3 is just for sure
		if((sayCoachCnt < serverParams->say_coach_cnt_max) &&
		   (continousTimeInPlayOn / ffwp > 0) &&
		   (continousTimeInPlayOn % ffwp > 3) &&
		   (continousTimeInPlayOn % ffwp < serverParams->freeform_wait_period - 3))
		{
			return 2;
		}
	}

	return 0;
}

void CoachKernel::SayMessage(const char* msg)
{
	static int i;

	if(msg == NULL) { // send the first message in FIFObuffer
		if(FIFObottom != FIFOtop) // buffer is not empty
		{
			if((i = CanSendMessage()) > 0)
			{
				if(i == 2) // we can send freeform message
				{
					sayCoachCnt++;
					// \todo we have to alter message to '(freeform "<message>")'
				}
				commandGen->Say(FIFObuffer[FIFObottom++]);
				if(FIFObottom > FIFO_BUFFER_SIZE) FIFObottom = 0;
				// +3 just to be sure
				canSendCLmessage = serverParams->clang_win_size + 3;
			}
		}
	} 
	else // msg != NULL
	{
		if((i = CanSendMessage()) > 0) // send it immediately
		{
			if(i == 2) // we can send freeform message
			{
				sayCoachCnt++;
				// \todo we have to alter message to '(freeform "<message>")'
			}
			commandGen->Say(msg);
			// +3 just to be sure
			canSendCLmessage = serverParams->clang_win_size + 3;
		}
		else // place it into the buffer
		{
			strcpy(FIFObuffer[FIFOtop++], msg);
			if(FIFOtop > FIFO_BUFFER_SIZE) FIFOtop = 0;
			if(FIFOtop == FIFObottom) // buffer is full
			{
				FIFObottom++;
				if(FIFObottom > FIFO_BUFFER_SIZE) FIFObottom = 0;
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// CoachKernel destruction

CoachKernel::~CoachKernel()
{
	int i, j;
	PLAYER_STATISTICS wholeTeam;

	DBGN("  Cycles in play on mode: %d", timeInPlayOn);
	DBGN("  Cycles ball on our side: %d", statsBall.cyclesOnOurHalf);
	DBGN("  Cycles ball on their side: %d", statsBall.cyclesOnTheirHalf);
	if(timeInPlayOn == 0) timeInPlayOn = 1;

	// store informations about movable objects position density to the file
	PlaygroundStatStoreToFile();

	memset(&wholeTeam, 0, sizeof(wholeTeam));
	for(i = 1; i <= 11; i++)
	{
		if(statsOurPlayer[i].cyclesNearBall == 0) 
			statsOurPlayer[i].cyclesNearBall = 1;

		DBGN("Our player number %2d:", i);
		DBGN(" - shoots on goal: %d", statsOurPlayer[i].shootsOnGoal);
		wholeTeam.shootsOnGoal += statsOurPlayer[i].shootsOnGoal;
		if(statsOurPlayer[i].shootsOnGoal > 0){
			DBGN(" - average length of shoot: %.2lf", statsOurPlayer[i].sumLenShootsOnGoal / statsOurPlayer[i].shootsOnGoal);
			wholeTeam.sumLenShootsOnGoal += statsOurPlayer[i].sumLenShootsOnGoal;
		}

		DBGN(" - cycles with ball: %d", statsOurPlayer[i].cyclesWithBall);
		wholeTeam.cyclesWithBall += statsOurPlayer[i].cyclesWithBall;

		DBGN(" - average distance from ball (x): %f", 
			statsOurPlayer[i].sumDistFromBall.GetX() / timeInPlayOn); 
		DBGN(" - average distance from ball (y): %f", 
			statsOurPlayer[i].sumDistFromBall.GetY() / timeInPlayOn);
		wholeTeam.sumDistFromBall += statsOurPlayer[i].sumDistFromBall;

		DBGN(" - average speed near ball: %f", 
			statsOurPlayer[i].sumSpeedNearBall / statsOurPlayer[i].cyclesNearBall);
		wholeTeam.sumSpeedNearBall += statsOurPlayer[i].sumSpeedNearBall;
		wholeTeam.cyclesNearBall += statsOurPlayer[i].cyclesNearBall;

		for(j = 0; j < PASS_LEN_COUNT; j++)
		{
			DBGN(" - number of all passes of length to %.1fm: %d", passLens[j], statsOurPlayer[i].passesAll[j]);
			wholeTeam.passesAll[j] += statsOurPlayer[i].passesAll[j];
		}
		DBGN(" - number of all passes of length more than %.1fm: %d", passLens[j - 1], statsOurPlayer[i].passesAll[j]);
		wholeTeam.passesAll[j] += statsOurPlayer[i].passesAll[j];

		for(j = 0; j < PASS_LEN_COUNT; j++)
		{
			DBGN(" - number of successful passes of length to %.1fm: %d", passLens[j], statsOurPlayer[i].passesSuccess[j]);
			wholeTeam.passesSuccess[j] += statsOurPlayer[i].passesSuccess[j];
			if(statsOurPlayer[i].passesSuccess[j] == 0) 
				statsOurPlayer[i].passesSuccess[j] = 1;
		}
		DBGN(" - number of successful passes of length more than %.1fm: %d", passLens[j - 1], statsOurPlayer[i].passesSuccess[j]);
		wholeTeam.passesSuccess[j] += statsOurPlayer[i].passesSuccess[j];

//		DBGN(" - average length of successful passes: %f", 
//			statsOurPlayer[i].sumPassSucLen / statsOurPlayer[i].passesSuccess);
//		wholeTeam.sumPassSucLen += statsOurPlayer[i].sumPassSucLen;

		DBGN(" - number of received passes: %d", statsOurPlayer[i].passesReceived);
		wholeTeam.passesReceived += statsOurPlayer[i].passesReceived;

		DBGN(" - number of interceptions: %d", statsOurPlayer[i].interceptions);
		wholeTeam.interceptions += statsOurPlayer[i].interceptions;

		wholeTeam.cyclesAttacker += statsOurPlayer[i].cyclesAttacker;
		wholeTeam.cyclesDefender += statsOurPlayer[i].cyclesDefender;
		wholeTeam.cyclesMidfielder += statsOurPlayer[i].cyclesMidfielder;
	}
	DBGN("------------------------------------\nSummary for our team:");
	DBGN(" - shoots on goal: %d", wholeTeam.shootsOnGoal);
	if(wholeTeam.shootsOnGoal > 0){
		DBGN(" - average length of shoot: %.2lf", wholeTeam.sumLenShootsOnGoal / wholeTeam.shootsOnGoal);
	}
	DBGN(" - cycles with ball: %d", wholeTeam.cyclesWithBall);
	DBGN(" - average speed near ball: %f", 
		wholeTeam.sumSpeedNearBall / wholeTeam.cyclesNearBall);
	for(j = 0; j < PASS_LEN_COUNT; j++)
	{
		DBGN(" - number of all passes of length to %.1fm: %d", passLens[j], wholeTeam.passesAll[j]);
		DBGN(" - number of successful passes of length to %.1fm: %d", passLens[j], wholeTeam.passesSuccess[j]);
	}
	DBGN(" - number of all passes of length more than %.1fm: %d", passLens[j - 1], wholeTeam.passesAll[j]);
	DBGN(" - number of successful passes of length more than %.1fm: %d", passLens[j - 1], wholeTeam.passesSuccess[j]);

	DBGN(" - %% of successful passes for various lengths respectively:");
	for(j = 0; j <= PASS_LEN_COUNT; j++)
	{
		if(wholeTeam.passesAll[j] == 0) TXT("N/A ");
		else TXT("%.2f ", (float)wholeTeam.passesSuccess[j] / wholeTeam.passesAll[j] * 100.0f);
	}
	DBGN("");

	DBGN(" - estimated number of defenders, midfielders and attackers, respectively:");
	DBGN("%.2f %.2f %.2f", 
		(float)wholeTeam.cyclesDefender / timeInPlayOn, 
		(float)wholeTeam.cyclesMidfielder / timeInPlayOn, 
		(float)wholeTeam.cyclesAttacker / timeInPlayOn);

	/*
	if(wholeTeam.passesSuccess == 0) 
		wholeTeam.passesSuccess = 1;
	DBGN(" - average length of successful passes: %f", 
		wholeTeam.sumPassSucLen / wholeTeam.passesSuccess);
*/
	DBGN(" - number of interceptions: %d", wholeTeam.interceptions);

	DBGN("------------------------------------");
	memset(&wholeTeam, 0, sizeof(wholeTeam));
	for(i = 1; i <= 11; i++)
	{
		if(statsTheirPlayer[i].cyclesNearBall == 0) 
			statsTheirPlayer[i].cyclesNearBall = 1;

		DBGN("Their player number %2d:", i);
		DBGN(" - shoots on goal: %d", statsTheirPlayer[i].shootsOnGoal);
		wholeTeam.shootsOnGoal += statsTheirPlayer[i].shootsOnGoal;
		if(statsTheirPlayer[i].shootsOnGoal > 0){
			DBGN(" - average length of shoot: %.2lf", statsTheirPlayer[i].sumLenShootsOnGoal / statsTheirPlayer[i].shootsOnGoal);
			wholeTeam.sumLenShootsOnGoal += statsTheirPlayer[i].sumLenShootsOnGoal;
		}

		DBGN(" - cycles with ball: %d", statsTheirPlayer[i].cyclesWithBall);
		wholeTeam.cyclesWithBall += statsTheirPlayer[i].cyclesWithBall;

		DBGN(" - average distance from ball (x): %f", 
			statsTheirPlayer[i].sumDistFromBall.GetX() / timeInPlayOn); 
		DBGN(" - average distance from ball (y): %f", 
			statsTheirPlayer[i].sumDistFromBall.GetY() / timeInPlayOn);
		wholeTeam.sumDistFromBall += statsTheirPlayer[i].sumDistFromBall;

		DBGN(" - average speed near ball: %f", 
			statsTheirPlayer[i].sumSpeedNearBall / statsTheirPlayer[i].cyclesNearBall);
		wholeTeam.sumSpeedNearBall += statsTheirPlayer[i].sumSpeedNearBall;
		wholeTeam.cyclesNearBall += statsTheirPlayer[i].cyclesNearBall;

		for(j = 0; j < PASS_LEN_COUNT; j++)
		{
			DBGN(" - number of all passes of length to %.1fm: %d", passLens[j], statsTheirPlayer[i].passesAll[j]);
			wholeTeam.passesAll[j] += statsTheirPlayer[i].passesAll[j];
		}
		DBGN(" - number of all passes of length more than %.1fm: %d", passLens[j - 1], statsTheirPlayer[i].passesAll[j]);
		wholeTeam.passesAll[j] += statsTheirPlayer[i].passesAll[j];

		for(j = 0; j < PASS_LEN_COUNT; j++)
		{
			DBGN(" - number of successful passes of length to %.1fm: %d", passLens[j], statsTheirPlayer[i].passesSuccess[j]);
			wholeTeam.passesSuccess[j] += statsTheirPlayer[i].passesSuccess[j];
			if(statsTheirPlayer[i].passesSuccess[j] == 0) 
				statsTheirPlayer[i].passesSuccess[j] = 1;
		}
		DBGN(" - number of successful passes of length more than %.1fm: %d", passLens[j - 1], statsTheirPlayer[i].passesSuccess[j]);
		wholeTeam.passesSuccess[j] += statsTheirPlayer[i].passesSuccess[j];

//		DBGN(" - average length of successful passes: %f", 
//			statsTheirPlayer[i].sumPassSucLen / statsTheirPlayer[i].passesSuccess);
//		wholeTeam.sumPassSucLen += statsTheirPlayer[i].sumPassSucLen;

		DBGN(" - number of received passes: %d", statsTheirPlayer[i].passesReceived);
		wholeTeam.passesReceived += statsTheirPlayer[i].passesReceived;

		DBGN(" - number of interceptions: %d", statsTheirPlayer[i].interceptions);
		wholeTeam.interceptions += statsTheirPlayer[i].interceptions;

		wholeTeam.cyclesAttacker += statsTheirPlayer[i].cyclesAttacker;
		wholeTeam.cyclesDefender += statsTheirPlayer[i].cyclesDefender;
		wholeTeam.cyclesMidfielder += statsTheirPlayer[i].cyclesMidfielder;
	}
	DBGN("------------------------------------\nSummary for their team:");
	DBGN(" - shoots on goal: %d", wholeTeam.shootsOnGoal);
	if(wholeTeam.shootsOnGoal > 0){
		DBGN(" - average length of shoot: %.2lf", wholeTeam.sumLenShootsOnGoal / wholeTeam.shootsOnGoal);
	}
	DBGN(" - cycles with ball: %d", wholeTeam.cyclesWithBall);
	DBGN(" - average speed near ball: %f", 
		wholeTeam.sumSpeedNearBall / wholeTeam.cyclesNearBall);
	for(j = 0; j < PASS_LEN_COUNT; j++)
	{
		DBGN(" - number of all passes of length to %.1fm: %d", passLens[j], wholeTeam.passesAll[j]);
		DBGN(" - number of successful passes of length to %.1fm: %d", passLens[j], wholeTeam.passesSuccess[j]);
	}
	DBGN(" - number of all passes of length more than %.1fm: %d", passLens[j - 1], wholeTeam.passesAll[j]);
	DBGN(" - number of successful passes of length more than %.1fm: %d", passLens[j - 1], wholeTeam.passesSuccess[j]);

	DBGN(" - %% of successful passes for various lengths respectively:");
	for(j = 0; j <= PASS_LEN_COUNT; j++)
	{
		if(wholeTeam.passesAll[j] == 0) TXT("N/A ");
		else TXT("%.2f ", (float)wholeTeam.passesSuccess[j] / wholeTeam.passesAll[j] * 100.0f);
	}
	DBGN("");

	DBGN(" - estimated number of defenders, midfielders and attackers, respectively:");
	DBGN("%.2f %.2f %.2f", 
		(float)wholeTeam.cyclesDefender / timeInPlayOn, 
		(float)wholeTeam.cyclesMidfielder / timeInPlayOn, 
		(float)wholeTeam.cyclesAttacker / timeInPlayOn);

	/*
	if(wholeTeam.passesSuccess == 0) 
		wholeTeam.passesSuccess = 1;
	DBGN(" - average length of successful passes: %f", 
		wholeTeam.sumPassSucLen / wholeTeam.passesSuccess);
*/
	DBGN(" - number of interceptions: %d", wholeTeam.interceptions);

	// simulation parameters
	if(serverParams != NULL) delete serverParams;
	if(playerParams != NULL) delete playerParams;
	if(playerTypes != NULL) delete[] playerTypes;
}

static int XpmReadFileToData(const char* name, char*** data)
{
	FILE* fin;
	char c, pc, r[1000];
	int cnt, lines;
	bool commentInline, commentLong, quote;
	
	fin = fopen(name, "rt");
	if(fin == NULL) return -1;
	*data = NULL;
	commentInline = false;
	commentLong = false;
	quote = false;
	lines = 0;
	c = '\0';

	while(1)
	{
		pc = c;
		if(fscanf(fin, "%c", &c) != 1) break;
		
		if(quote)
		{
			if(c == '"' && pc != '\\')
			{
				quote = false;
				r[cnt] = '\0';
				lines++;

				*data = (char**) realloc(*data, lines * sizeof(char*));
				if(*data == NULL){fclose(fin);return -2;}

				(*data)[lines - 1] = (char*) malloc((cnt + 1) * sizeof(char));
				if((*data)[lines - 1] == NULL){fclose(fin);return -2;}

				strcpy((*data)[lines - 1], r);
			}
			else
			{
				if(c == '"' && pc == '\\') r[cnt - 1] = '"';
				else r[cnt++] = c;
			}
		}
		else if(commentLong)
		{
			if(c == '/' && pc == '*') commentLong = false;
		}
		else if(commentInline)
		{
			if(c == '\n') commentInline = false;
		}
		else
		{
			if(c == '"') {quote = true; cnt = 0;}
			if(c == '/' && pc == '/') commentInline = true;
			if(c == '*' && pc == '/') commentLong = true;
		}
	}

	fclose(fin);

	return lines;
}

static void XpmFree(char*** data, int lines)
{
	static int i;
	for(i = 0; i < lines; i++)
	{
		free((*data)[i]);
		(*data)[i] = NULL;
	}
	free(*data);
	*data = NULL;
}

bool CoachKernel::SendTeamGraphic(const char* path)
{
	int i, j, k;
	char name[512];
	char msg[2048];
	char** data;

	// If we have already sent all graphics, do nothing
	if(teamGraphicOkCount >= 256) return true;
	teamGraphicTries++;

	for(i = 0; i < 32; i++)
	{
		for(j = 0; j < 8; j++)
		{
			if(!teamGraphicOk[i][j])
			{
			    sprintf(name, "%slogo-%02i-%02i.xpm", path, i, j);
    
			    int rval = XpmReadFileToData(name, &data);
			    if(rval <= 0) return false;
    
			    int r, c, p, a;
			    if(sscanf(data[0], "%i %i %i %i", &r, &c, &p, &a) != 4) return false;
			    int lines = r + p + a;
    
			    msg[0] = '\0';
			    for(k = 0; k < lines; k++)
			    {
				    if(k != 0) strcat(msg, " ");
				    strcat(msg, "\"");
				    strcat(msg, data[k]);
				    strcat(msg, "\"");
			    }
    
			    XpmFree(&data, rval);
    
				if(msg[0] != '\0') (void) commandGen->TeamGraphic(i, j, msg);
				else return false;
			}
		}
	}

	return true;
}

/// Handler for 'ok team_graphic' messages
void CoachKernel::OnOkTeamGraphic(int x, int y)
{
	teamGraphicOk[x][y] = true;
	teamGraphicOkCount++;
}

void CoachKernel::CoachBeforeKikcOff()
{
	static bool bWasChanged=false;
	if (!bWasChanged){
		int id=-1;
		int idMiddle=-1;
		float dash_p_r=-1.0;
		float stamina_inc_max=-1.0;
		float realSpeed = -1.0;
		for (int i=0; i<playerParams->player_types; i++){
			float curRealSpeedMax = CalcRealSpeedMax(i) ;
			if (realSpeed<curRealSpeedMax){
				realSpeed = curRealSpeedMax;
				stamina_inc_max = playerTypes[i].stamina_inc_max;
				idMiddle = id;
				id=i;
			}
			else if (realSpeed==curRealSpeedMax && stamina_inc_max<playerTypes[i].stamina_inc_max)
			{
				realSpeed = curRealSpeedMax;
				stamina_inc_max = playerTypes[i].stamina_inc_max;
				idMiddle = id;
				id=i;
			}
		}
		
		commandGen->ChangePlayerType(9, id);	
		commandGen->ChangePlayerType(10, id);	
		commandGen->ChangePlayerType(11, id);	
		if (idMiddle!=-1)
		{
			commandGen->ChangePlayerType(5, idMiddle);	
			commandGen->ChangePlayerType(8, idMiddle);	
		}

		
		bWasChanged=true;
	}
}


void CoachKernel::CouldChangePlayerType()
{
	
}

void CoachKernel::Run(int iTask)
{
	if (playerParams!=NULL && playerParams->player_types>0 &&nHeteroCounter==playerParams->player_types){
		if (GetPlayMode()==PM_BeforeKickOff){
			CoachBeforeKikcOff();
			return;
		}
		else {
			if (IsTimeStoped()){
					CouldChangePlayerType();	
			}
		}
	}
}

/**
 * Send our players information about oponet heterotypes 
 *
 *  \author Eugen Krizo
 *
 */
void CoachKernel::SendOpponentHeterotypes()
{

	char msg[1000];
	sprintf(msg, "(freeform \"(opponentHeterotypes ");
	int nCount = 0;
	for (int i=0; i<12; i++)
	{
		if (spyTheirPlayer[i].heteroType>0)
		{
			sprintf(msg, "%s(%d %d)", msg,i, spyTheirPlayer[i].heteroType);
			nCount++;
		}
	}
	sprintf(msg, "%s)\")", msg);
	printf("\n%s",msg); 
	
	if (nCount>0) SayMessage(msg);
}

/**
 * Whether time specified for the period of the storing the statistics elapsed
 *
 * \author Marian Tines
 *
 */
bool CoachKernel::PlaygroundStatIsTimeToStore()
{
	if (statsPlaygroundStorePeriod <= 0)
		return false;

	if (GetSimTime() - statsPlaygroundStoreLastSimTime > statsPlaygroundStorePeriod)
		return true;

	return false;
}

/**
 * Stores playground statistics to file
 *
 * \author Marian Tines
 *
 */
void CoachKernel::PlaygroundStatStoreToFile()
{
	char buffer[80];
	time_t	timeUnix;
	tm		timeLocal;
	time(&timeUnix);
	timeLocal = *localtime(&timeUnix);
	sprintf(buffer, "PlaygroundStat_%04d%02d%02d%02d%02d-%d.log", 1900 + timeLocal.tm_year, 1 + timeLocal.tm_mon, timeLocal.tm_mday, timeLocal.tm_hour, timeLocal.tm_min, GetSimTime());
	statsPlayground->SerializeYourself(buffer, "Debug/PlayerStat");

	statsPlaygroundStoreLastSimTime = GetSimTime();
}

// format freeform  "(10.5 9 D)(19.9 9.4 F)..."
// x, y, rola
	void CoachKernel::NewForm(Point newPos, char* role)
	{

	if(strlen(NewFormationBuffer) > 100)
	{
		SayNewForm();
	}
	else
	{
		char buffer[128];
		strcat(NewFormationBuffer,"(");
		sprintf(buffer, "%G", newPos.GetX());
		strcat(NewFormationBuffer,buffer);
		strcat(NewFormationBuffer," ");
		sprintf(buffer, "%G", newPos.GetY());
		strcat(NewFormationBuffer,buffer);
	
		

		if(role != NULL)
		{
			strcat(NewFormationBuffer," ");
			if(strcmp(role,"PR_Defensman")==0)
			{
				strcat(NewFormationBuffer,"D");
				
			}
			else if(strcmp(role,"PR_Midfielder")==0)
			{
				strcat(NewFormationBuffer,"M");
				
			}
			else if(strcmp(role,"PR_Forward")==0)
			{
				strcat(NewFormationBuffer,"F");
				
			}
		}
		strcat(NewFormationBuffer,")");
	}

	}
// format freeform  "(1 10.5 9 D)(8 19.9 9.4 F)..."
// cislo hraca, x, y, rola
	void CoachKernel::ChangePlayerPos(int playerNum, Point newPos, char* role)
	{

	if(strlen(FormationBuffer) > 100)
	{
		SayNewPos();
	}
	else
	{
		char buffer[128];
		strcat(FormationBuffer,"(");
		sprintf(buffer, "%i", playerNum);
		strcat(FormationBuffer,buffer);
		strcat(FormationBuffer," ");
		sprintf(buffer, "%G", newPos.GetX());
		strcat(FormationBuffer,buffer);
		strcat(FormationBuffer," ");
		sprintf(buffer, "%G", newPos.GetY());
		strcat(FormationBuffer,buffer);

		currentFormation.SetPlayerPosition(0,playerNum,newPos);
	
		if(role != NULL)
		{
			strcat(FormationBuffer," ");
			if(strcmp(role,"PR_GoalKeeper")==0)
			{
				strcat(FormationBuffer,"G");
				currentFormation.SetPlayerRole(0,playerNum,PR_GoalKeeper);
			}
			else if(strcmp(role,"PR_Defensman")==0)
			{
				strcat(FormationBuffer,"D");
				currentFormation.SetPlayerRole(0,playerNum,PR_Defensman);
			}
			else if(strcmp(role,"PR_Midfielder")==0)
			{
				strcat(FormationBuffer,"M");
			currentFormation.SetPlayerRole(0,playerNum,PR_Midfielder);
			}
			else if(strcmp(role,"PR_Forward")==0)
			{
				strcat(FormationBuffer,"F");
				currentFormation.SetPlayerRole(0,playerNum,PR_Forward);
			}
		}
		strcat(FormationBuffer,")");
	}

	}
	void CoachKernel::ChangeForm (int formation)
	{
		char msg[128];
		char buffer[3];
		strcpy(msg,"(freeform \"(formation ");
		sprintf(buffer, "%i", formation);

		strcat(msg,buffer);
		strcat(msg,")\")");
		SayMessage(msg);
		printf("\nDEBUG: %s",msg);
	}
	void CoachKernel::SayNewPos ()
	{
		if(strcmp(FormationBuffer,"") != 0)
		{
		FILE *fp = fopen("saydbg.txt", "w");
		fprintf(fp, "\naaaaa: %s",FormationBuffer);
		char msg[128];
		strcpy(msg,"(freeform \"(ChPl ");
		strcat(msg,FormationBuffer);
		strcat(msg,")\")");
		fprintf(fp, "\nDEBUG: %s",msg);
		printf("\naaaaa: %s",msg);
		SayMessage(msg);
		strcpy(FormationBuffer,"");
		fclose(fp);
		}
	}
	
	void CoachKernel::SayNewForm ()
	{
		//if(strcmp(NewFormationBuffer,"") != 1)
		//{
		FILE *fp = fopen("saydbg.txt", "a");
		fprintf(fp, "\naaaaa: %s",NewFormationBuffer);
		
		char msg[128];
		strcpy(msg,"(freeform \"(NF ");
		strcat(msg,NewFormationBuffer);
		strcat(msg,")\")");
		fprintf(fp, "\nDEBUG: %s",msg);
		SayMessage(msg);
		printf("\naaaaa: %s",msg);
		strcpy(NewFormationBuffer,"");
		fclose(fp);
	//}
	}

	int CoachKernel::countCoverage(BitPlaygroundObjectStat* bitStat, int x_pos, int y_pos) {
		int count = 0;
		for (int i=0; i<3;i++) 
			for (int j=0; j<3; j++)
				if ((bitStat->GetBitPlaygroundStat(x_pos-1+i, y_pos-1+j)) && (x_pos > -1*HALF_FIELD_LENGTH) && (x_pos < HALF_FIELD_LENGTH) &&
					(y_pos > -1*HALF_FIELD_WIDTH) && (y_pos < HALF_FIELD_WIDTH)) count++;
		return count;
	}
	
	void CoachKernel::ChangeAbsoluteFormation(int def, int mid, int att) {
	//using bitmap of coverage changing formation

		//FILE *fp = fopen("mydbg.txt", "a");
		//fprintf(fp, "Je cas %d, skore je %d:%d a menim formaciu na %d - %d - %d \n", timeInPlayOn, scoreOur, scoreTheir, def, mid, att);
		
		//if (ourSide) fprintf(fp, "Hrame vlavo\n");
		//else fprintf(fp, "Hrame vpravo\n");

		float swapSides;
		if (ourSide) swapSides = 1; 
		else swapSides = -1;

		time_t cl_start, cl_end;
		time(&cl_start);
		
		BitPlaygroundObjectStat* myBitStat = new BitPlaygroundObjectStat();
		myBitStat->LoadByPeakAlgorithm(statsPlayground, 10, ENEMY);

		int i,j,k;
		Point max_point, abs_position;
		float tmp_val, max_heur;
		
		
		for (i=0; i<def; i++) {	
			abs_position = currentFormation.GetPlayerPositionInFormation(i+2, currentFormation.GetFormationId(def, mid, att));
			abs_position.SetX(abs_position.GetX()*swapSides);
			max_heur = -1000;
			for (j=0;j<9;j++) {
				for (k=0;k<9;k++) {
					tmp_val = (float) countCoverage(myBitStat, abs_position.GetX()-4+k, abs_position.GetY()-4+j);
					tmp_val = tmp_val - (float) (0.3*((abs(5-k) + abs(5-j))));
					if (tmp_val > max_heur) {
						max_point = Point(abs_position.GetX()-4+k, abs_position.GetY()-4+j);
						max_heur = tmp_val;
					}
				}
			}
			if (max_heur == 0) max_point = abs_position;
			//NewForm(max_point, "PR_Defensman");
			//fprintf (fp, "hrac c %d na poziciu x= %f, y= %f \n", i+2, max_point.GetX(), max_point.GetY());
		}

		for (i=0; i<mid; i++) {	
			abs_position = currentFormation.GetPlayerPositionInFormation(i+def+2, currentFormation.GetFormationId(def, mid, att));
			abs_position.SetX(abs_position.GetX()*swapSides);
			max_heur = -1000;
			for (j=0;j<9;j++) {
				for (k=0;k<9;k++) {
					tmp_val = (float) countCoverage(myBitStat, abs_position.GetX()-4+k, abs_position.GetY()-4+j);
					tmp_val = tmp_val - (float) (0.3*((abs(5-k) + abs(5-j))));
					if (tmp_val > max_heur) {
						max_point = Point(abs_position.GetX()-4+k, abs_position.GetY()-4+j);
						max_heur = tmp_val;
					}
				}
			}
			if (max_heur == 0) max_point = abs_position;
			//NewForm(max_point, "PR_Midfielder");
			//fprintf (fp, "hrac c %d na poziciu x= %f, y= %f \n", i+def+2, max_point.GetX(), max_point.GetY());
		}

		for (i=0; i<att; i++) {	
			abs_position = currentFormation.GetPlayerPositionInFormation(i+def+mid+2, currentFormation.GetFormationId(def, mid, att));
			abs_position.SetX(abs_position.GetX()*swapSides);
			max_heur = 1000;
			for (j=0;j<9;j++) {
				for (k=0;k<9;k++) {
					tmp_val = (float) countCoverage(myBitStat, abs_position.GetX()-4+k, abs_position.GetY()-4+j);
					tmp_val = tmp_val - (float) (0.3*((abs(5-k) + abs(5-j))));
					if (tmp_val < max_heur) {
						max_point = Point(abs_position.GetX()-4+k, abs_position.GetY()-4+j);
						max_heur = tmp_val;
					}
				}
			}
			if (max_heur == 1000) max_point = abs_position;
			//NewForm(max_point, "PR_Forward");
			//fprintf (fp, "hrac c %d na poziciu x= %f, y= %f \n", i+def+mid+2, max_point.GetX(), max_point.GetY());
		}
		//fprintf(fp, "Teraz som vykonal setko co bolo treba idem uz len volat SayNewPos, ak to padne chyba je u votrelca\n");
		//SayNewForm();
		formationLastChangeTime = timeInPlayOn;
		delete myBitStat;
		time(&cl_end);
		//fprintf (fp, "Exec time  %6.2f seconds\n", cl_end - cl_start);

		//fclose(fp);

	}

	//FINITO MASTER BOYS