/**
 * \file Ball.cpp
 * \date 12.04.2003
 *
 * Interface for the Ball class
 */

#include "Player/World/Ball.h"
#include "Player/World/World.h"
#include <vector>

Ball::Ball(World* w) : Movable(w)
{
	valid = false;
	lastSeenTime = -2;
}

Ball::Ball(const Ball& ball) : Movable(ball)
{
	valid = ball.valid;
	lastSeenPos = ball.lastSeenPos;
	lastSeenTime = ball.lastSeenTime;
}

Ball& Ball::operator=(const Ball& ball)
{
	if(&ball == this) return *this;
	Movable::operator=(ball);

	valid = ball.valid;
	lastSeenPos = ball.lastSeenPos;
	lastSeenTime = ball.lastSeenTime;

	return *this;
}

/// \brief Prepocita polohu lopty z vizualnej informacie,
///      \li ak je lopta vo vizualnej informacii tak ju urci na zaklade nej (vidi loptu)
///		 \li inak ju predikuje 
/// \param sb Sense body
/// \param vi Vizualna informacia
void Ball::Recalculate(const SENSE_BODY* sb, const VISUAL_INFO* vi)
{		
	// I am looking this way
	Angle viewDir = world->me->GetAbsDir() + world->me->GetNeckDir();

	// I see a ball and I know its distance and direction, compute its absolute position
	Angle angle = viewDir + DegAngle(vi->ball.dir);
	pos = world->me->pos + PolarVector(vi->ball.dist, angle);

	WORLD_RECALC_DBG("  ball: viewDir: %f, dir: %f", viewDir.GetDeg(), vi->ball.dir);

	// Do I see also its speed?
	if(vi->ball.changesValid)
	{
		// From TsinghuAeolus2002
		Vector rpos = PolarVector(vi->ball.dist, DegAngle(vi->ball.dir));
		rpos /= rpos.GetLength();

		Vector rvel = RectVector(
			vi->ball.distChange * rpos.GetX() - vi->ball.dirChange * M_PI / 180.f * vi->ball.dist * rpos.GetY(),
			vi->ball.distChange * rpos.GetY() + vi->ball.dirChange * M_PI / 180.f * vi->ball.dist * rpos.GetX());
			
		// Rotate rvel
		rvel += viewDir;
		vel = world->me->vel + rvel;

//		vel = world->me->vel + PolarVector(vi->ball.distChange, angle) + 
//			PolarVector(vi->ball.dist * dirCh, angle).OrthoLeft();

		WORLD_RECALC_DBG("  ball: changesValid, pos: (%f, %f), vel: (%f, %f)", pos.GetX(), pos.GetY(), vel.GetX(), vel.GetY());
	}
	else
	{
		// I do not see ball speed, use the CMU method
		if((vi->ball.dist <= world->params->server->visible_distance) &&
			(world->state->GetSimTime() == lastSeenTime + 1) &&
			(world->state->GetSimTime() == world->me->GetLastSeenTime() + 1))
		{
			Vector correction = 
				world->me->vel / world->me->type->player_decay - 
				(world->me->pos - world->me->GetLastSeenPos());
				
			vel = (pos - lastSeenPos + correction);
			vel *= world->params->server->ball_decay;
		}

		// On collision, multiply ball speed by -0.1
		bool ballCollision = false;
		if(vi->ball.dist <= world->me->type->player_size + world->params->server->ball_size + 0.05)
		{
			ballCollision = true;
		}

		if(ballCollision && (world->state->GetSimTime() > lastSeenTime)) vel *= -0.1f;

		WORLD_RECALC_DBG("  ball: CMU method, pos: (%f, %f), vel: (%f, %f)", pos.GetX(), pos.GetY(), vel.GetX(), vel.GetY());
	}

	// I see the ball just now
	age = 0;
	lastSeenPos = pos;
	lastSeenTime = world->state->GetSimTime();
	valid = true;
}

/// \brief Predikuje polohu a rychlost lopty o jeden simulacny krok dopredu,
void Ball::Step()
{
	pos = pos + vel;
	vel *= world->params->server->ball_decay;
	age++;
}

/// \brief Predikuje pohyb lopty na nStep krokov dopredu
/// \param nStep Pocet krokov, na ktor ma predikovat pohyb lopty
/// \param ballVector Vektor lopt, kam sa ukladaju informacie o stave lopty v danom ktoku
///        ballVector[0] = stav lopty po 1 somulacnom kroku == Ball::Step
void Ball::Predict(int nStep, BALLARRAY& ballVector)const
{
	ballVector.clear();
	Ball b(*this);
	for (int i=0; i<nStep; i++){
		b.Step();
		ballVector.push_back(b);
	}
}

/// \brief Predikuje pohyb lopty na nStep krokov dopredu ak jej bola udelena pociatocna rychlost initVel
/// \param nStep Pocet krokov, na ktor ma predikovat pohyb lopty
/// \param initVel Udelena rychlost lopte
/// \param ballVector Vektor lopt, kam sa ukladaju informacie o stave lopty v danom ktoku
///        ballVector[0] = stav lopty po 1 somulacnom kroku == Ball::Step
void Ball::Predict(int nStep, Vector accel, BALLARRAY& ballVector)const
{
	ballVector.clear();
	Ball b(*this);
	if (b.vel.GetLength()<=0.05)
		b.vel.SetAngle(0.0);
	b.vel += accel;
	if (b.vel.GetLength()>world->params->server->ball_speed_max)
		b.vel.SetLength(world->params->server->ball_speed_max);

	for (int i=0; i<nStep; i++)
	{
		b.pos = b.pos + b.vel;
		b.vel = b.vel * world->params->server->ball_decay;
		ballVector.push_back(b);
	}
}

Ball::~Ball()
{
}
